@@ -21,22 +21,34 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
  */
 
+#include <unistd.h>
 #include <aconf.h>
 #include <math.h>
 #include "gpdf-view.h"
 #include "gpdf-util.h"
 #include "gpdf-marshal.c"
 #include "gpdf-g-switch.h"
+#  include <glib/gi18n.h>
 #  include <gdk/gdkkeysyms.h>
 #  include <libgnome/gnome-macros.h>
 #  include <libgnomeprint/gnome-print.h>
 #  include <libgnomeprintui/gnome-print-preview.h>
+#  include <libgnomevfs/gnome-vfs-mime.h>
+#  include <libgnomevfs/gnome-vfs-mime-handlers.h>
+#  include <gtk/gtkwidget.h>
+#  include <ggv-document.h>
 #  include "gtkgesture.h"
+#  include <glade/glade.h>
 #include "gpdf-g-switch.h"
+#include "gpdf-control.h"
+#include "gpdf-control-private.h"
 #include "GPOutputDev.h"
 #include "PDFDoc.h"
 #include "Link.h"
 #include "gpdf-links-canvas-layer.h"
+#ifdef USE_ANNOTS_VIEW
+#include "gpdf-annots-view.h"
+#endif
 
 BEGIN_EXTERN_C
 
@@ -58,14 +70,23 @@ struct _GPdfViewPrivate {
 	guint dragging : 1;
 	gint drag_anchor_x, drag_anchor_y;
 	gint drag_offset_x, drag_offset_y;
+
+	G_List *forward_history;
+	G_List *backward_history;
+
+#ifdef USE_ANNOTS_VIEW
+	gboolean (*annots_cb)(Annot *annot, gpointer data);
+	GPdfAnnotsView *annots_view;
+#endif
+	GPdfControl *parent;
 };
 
 #define PARENT_TYPE GNOME_TYPE_CANVAS
-GPDF_CLASS_BOILERPLATE (GPdfView, gpdf_view, GnomeCanvas, PARENT_TYPE);
+static GnomeCanvasClass *parent_class = NULL;
 
 #define GPDF_IS_NON_NULL_VIEW(obj) (((obj) != NULL) && (GPDF_IS_VIEW ((obj))))
 #define GPDF_IS_VIEW_WITH_DOC(obj) \
-(GPDF_IS_NON_NULL_VIEW (obj) && (obj)->priv->pdf_doc)
+(GPDF_IS_NON_NULL_VIEW (obj) && ((GPdfView *)(obj))->priv->pdf_doc)
 
 enum {
 	ZOOM_CHANGED_SIGNAL,
@@ -78,7 +99,7 @@ enum {
 static guint gpdf_view_signals [LAST_SIGNAL];
 
 
-#define PAGE_PAD 3
+#define PAGE_PAD ((gdouble)3.0)
 
 static void
 gpdf_view_page_background_show (GPdfView *gpdf_view)
@@ -99,7 +120,7 @@ gpdf_view_page_background_hide (GPdfView *gpdf_view)
 }
 
 static void
-gpdf_view_page_background_resize (GPdfView *gpdf_view, 
+gpdf_view_page_background_resize (GPdfView *gpdf_view,
 				  gdouble width, gdouble height)
 {
 	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
@@ -112,10 +133,26 @@ gpdf_view_page_background_resize (GPdfView *gpdf_view,
 			       NULL);
 }
 
+#ifdef USE_ANNOTS_VIEW
+static GPdfAnnotsView *
+gpdf_view_get_annots_view (GPdfView *gpdf_view)
+{
+	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view),
+			      NULL);
+
+	if (!gpdf_view->priv->annots_view)
+	  gpdf_view->priv->annots_view = GPDF_ANNOTS_VIEW 
+	    (gpdf_control_private_get_annots_view (gpdf_view->priv->parent)); 
+
+	return gpdf_view->priv->annots_view; 
+}
+#endif
+
 static void
-setup_page_transform (gdouble *transform, 
-		      gdouble x1, gdouble y1, gdouble x2, gdouble y2,
-		      gint angle)
+gpdf_view_setup_page_transform (gdouble *transform,
+				gdouble x1, gdouble y1,
+				gdouble x2, gdouble y2,
+				gint angle)
 {
 	double width;
 	double height;
@@ -146,10 +183,10 @@ setup_page_transform (gdouble *transform,
 
 	art_affine_translate (translate, -x1, -y1);
 	art_affine_multiply (transform, translate, transform);
-}	
+}
 
 static gint
-canonical_multiple_of_90 (gint n)
+gpdf_view_canonical_multiple_of_90 (gint n)
 {
 	while (n >= 360) {
 		n -= 360;
@@ -161,6 +198,62 @@ canonical_multiple_of_90 (gint n)
 	return 90 * (gint)((n / 90.0) + .5);
 }
 
+static void
+gpdf_free_history_node (gpointer node, gpointer user_data)
+{
+	g_free (node);
+}
+
+static void
+gpdf_view_history_stack_page (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv = gpdf_view->priv;
+	GPdfViewHistory *hnode;
+
+	/* First delete forward history if one exists */
+	if (priv->forward_history) {
+		g_list_foreach (priv->forward_history,
+				gpdf_free_history_node,
+				NULL);
+		g_list_free (priv->forward_history);
+		priv->forward_history = (G_List *)NULL;
+	}
+
+	/* Prepend history node */
+	hnode = (GPdfViewHistory *)g_malloc (sizeof(GPdfViewHistory));
+	hnode->zoom = priv->zoom;
+	hnode->page = priv->current_page;
+	hnode->hval = gtk_adjustment_get_value(GTK_LAYOUT (gpdf_view)->hadjustment); 
+	hnode->vval = gtk_adjustment_get_value(GTK_LAYOUT (gpdf_view)->vadjustment); 
+	priv->backward_history = g_list_prepend (priv->backward_history, (gpointer)hnode);
+}
+
+void
+gpdf_view_clear_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	priv = gpdf_view->priv;
+
+	if (priv->forward_history) {
+		g_list_foreach (priv->forward_history,
+				gpdf_free_history_node,
+				NULL);
+		g_list_free (priv->forward_history);
+		priv->forward_history = NULL;
+	}
+
+	if (priv->backward_history) {
+		g_list_foreach (priv->backward_history,
+				gpdf_free_history_node,
+				NULL);
+		g_list_free (priv->backward_history);
+		priv->backward_history = NULL;
+	}	
+}
+
 static void
 gpdf_view_link_action_goto (GPdfView *gpdf_view, LinkGoTo *link_goto)
 {
@@ -196,9 +289,11 @@ gpdf_view_link_action_goto (GPdfView *gpdf_view, LinkGoTo *link_goto)
 
 		delete dest;
 
-		if (page != gpdf_view->priv->requested_page)
+		if (page != gpdf_view->priv->requested_page) {
 			gpdf_view_goto_page (gpdf_view, page);
-	}	
+			gpdf_view_scroll_to_top (gpdf_view);
+		}
+	}
 }
 
 static void
@@ -210,7 +305,7 @@ gpdf_view_link_action_named (GPdfView *gpdf_view, LinkNamed *link_named)
 	g_return_if_fail (dynamic_cast <LinkNamed *> (link_named) != NULL);
 
 	action_name = link_named->getName ()->getCString ();
-	
+
 	if (strcmp (action_name, "NextPage") == 0) {
 		gpdf_view_page_next (gpdf_view);
 	} else if (strcmp (action_name, "PrevPage") == 0) {
@@ -230,8 +325,83 @@ gpdf_view_link_action_named (GPdfView *gpdf_view, LinkNamed *link_named)
 }
 
 static void
-link_clicked_cb (GPdfLinksCanvasLayer *links_layer, Link *link,
-		 gpointer user_data)
+gpdf_view_link_action_uri (GPdfView *gpdf_view, LinkURI *link_uri)
+{
+	GString *uri; 
+	GnomeVFSMimeAction *action; 
+	GnomeVFSMimeApplication *app;
+	GnomeVFSURI *vfsuri; 
+	G_List *uris = NULL;
+	const gchar *mime;
+
+	g_message ("gpdf_view_link_action_uri: enters ..."); 
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	g_return_if_fail (dynamic_cast <LinkURI *> (link_uri) != NULL);
+
+	uri = link_uri->getURI (); 
+	g_message ("gpdf_view_link_action_uri: uri = '%s'", (gchar *)uri->getCString ()); 
+	uris = g_list_append (uris, (gchar *)uri->getCString ());
+
+	vfsuri = gnome_vfs_uri_new ((const gchar *)uris->data); 
+	mime = gnome_vfs_get_mime_type_common (vfsuri);
+
+	g_message ("gpdf_view_link_action_uri: mime = '%s'", mime); 
+
+	if (!mime) {
+		mime = "text/html";
+		g_message ("gpdf_view_link_action_uri: using default mime = '%s'", mime); 
+	}
+
+	action = gnome_vfs_mime_get_default_action (mime); 
+	if (gnome_vfs_mime_action_launch (action, uris) != GNOME_VFS_OK)
+	{
+		app = gnome_vfs_mime_get_default_application (mime); 
+		
+		if (!app || gnome_vfs_mime_application_launch (app, uris) != GNOME_VFS_OK) {
+
+			/* Fallback: No component or app was found. Use a browser */
+			mime = "text/html";
+			action = gnome_vfs_mime_get_default_action (mime);
+			if (gnome_vfs_mime_action_launch (action, uris) != GNOME_VFS_OK)
+			{
+				app = gnome_vfs_mime_get_default_application (mime); 
+				if (!app || gnome_vfs_mime_application_launch (app, uris) != GNOME_VFS_OK) {
+			
+					gchar *msg = 
+					  g_strdup_printf(_("Failed to launch %s for URI: %s"),
+							  app->name,
+							  (const gchar *)uris->data);
+					gpdf_control_private_error_dialog (gpdf_view->priv->parent, 
+									   _("Application Launch Failure !"),
+									   msg,
+									   TRUE);
+					g_free (msg);
+				}
+			}
+		}
+		else {
+			gchar *msg = g_strdup_printf ("URI '%s' launched with %s",
+						      (const gchar *)uris->data,
+						      app->name);
+			gpdf_control_private_info_dialog (gpdf_view->priv->parent, 
+							  _("Application Launch nottification !"),
+							  msg,
+							  TRUE);
+			g_free (msg);			
+		}
+
+		gnome_vfs_mime_application_free (app); 
+	}
+	g_free (uris); 
+	gnome_vfs_uri_unref (vfsuri); 
+	delete uri; 
+}
+
+static void
+gpdf_view_link_clicked_cb (GPdfLinksCanvasLayer *links_layer,
+			   Link                 *link,
+			   gpointer              user_data)
 {
 	GPdfView *gpdf_view;
 	LinkAction *link_action;
@@ -251,6 +421,10 @@ link_clicked_cb (GPdfLinksCanvasLayer *links_layer, Link *link,
 		gpdf_view_link_action_named (
 			gpdf_view, dynamic_cast <LinkNamed *> (link_action));
 		break;
+	case actionURI:
+		gpdf_view_link_action_uri (
+			gpdf_view, dynamic_cast <LinkURI *> (link_action));
+		break; 
 	default:
 		g_message ("Click on link: unimplemented Action");
 		break;
@@ -279,6 +453,8 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 	if (priv->requested_page == priv->current_page)
 		return FALSE;
 
+	gpdf_control_private_set_wait_cursor (priv->parent);
+	
 	page = priv->requested_page;
 
 	if (priv->print_context) {
@@ -290,11 +466,11 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 	pdf_page = pdf_doc->getCatalog ()->getPage (page);
 	rect = pdf_page->getBox ();
 
-	page_rotate = canonical_multiple_of_90 (pdf_doc->getPageRotate (page));
+	page_rotate = gpdf_view_canonical_multiple_of_90 (pdf_doc->getPageRotate (page));
 
-	setup_page_transform (page_transform,
-			      rect->x1, rect->y1, rect->x2, rect->y2,
-			      page_rotate);
+	gpdf_view_setup_page_transform (page_transform,
+					rect->x1, rect->y1, rect->x2, rect->y2,
+					page_rotate);
 
 	if (page_rotate == 90 || page_rotate == 270) {
 		page_width_rot = pdf_doc->getPageHeight (page);
@@ -318,11 +494,24 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 	gnome_print_config_unref (config);
 
 	priv->output_dev->setPrintContext (priv->print_context);
+#ifdef USE_ANNOTS_VIEW
 	priv->pdf_doc->displayPage (priv->output_dev,
 				    page,
 				    72,
 				    0,
-				    gFalse);
+				    gFalse,
+				    NULL, NULL, 
+				    priv->annots_cb,
+				    gpdf_view_get_annots_view (gpdf_view));
+#else
+	priv->pdf_doc->displayPage (priv->output_dev,
+				    page,
+				    72,
+				    0,
+				    gFalse,
+				    NULL, NULL, 
+				    NULL, NULL);
+#endif
 
 	priv->current_page = page;
 
@@ -336,7 +525,7 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 		gnome_canvas_root (GNOME_CANVAS (gpdf_view)),
 		GPDF_TYPE_LINKS_CANVAS_LAYER,
 		"links",
-		new Links (pdf_page->getAnnots (&obj), 
+		new Links (pdf_page->getAnnots (&obj),
 			   pdf_doc->getCatalog ()->getBaseURI ()),
 		NULL));
 	g_object_ref (G_OBJECT (priv->links_layer));
@@ -347,9 +536,13 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 	gnome_canvas_item_affine_relative (
 		GNOME_CANVAS_ITEM (priv->links_layer), page_transform);
 
-	g_signal_connect (G_OBJECT (priv->links_layer), "link_clicked",
-			  G_CALLBACK (link_clicked_cb), gpdf_view);
+	gnome_canvas_item_raise_to_top (GNOME_CANVAS_ITEM (priv->links_layer));
 
+	g_signal_connect (G_OBJECT (priv->links_layer), "link_clicked",
+			  G_CALLBACK (gpdf_view_link_clicked_cb),
+			  gpdf_view);
+
+	gpdf_control_private_reset_cursor (priv->parent);
 	return FALSE;
 }
 
@@ -361,8 +554,8 @@ gpdf_view_emit_page_changed (GPdfView *gpdf_view, gint page)
 		       0, page);
 }
 
-void
-gpdf_view_goto_page (GPdfView *gpdf_view, gint page)
+static void
+gpdf_view_goto_page_no_history (GPdfView *gpdf_view, gint page)
 {
 	GPdfViewPrivate *priv;
 	gint last_page;
@@ -375,12 +568,57 @@ gpdf_view_goto_page (GPdfView *gpdf_view, gint page)
 	g_idle_add ((GSourceFunc)gpdf_view_render_page, gpdf_view);
 }
 
+void
+gpdf_view_goto_page (GPdfView *gpdf_view, gint page)
+{
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+
+	gpdf_view_history_stack_page (gpdf_view);
+	gpdf_view_goto_page_no_history (gpdf_view, page);
+}
+
+void
+gpdf_view_scroll_to_top (GPdfView *gpdf_view)
+{
+	gint x, y;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (gpdf_view), &x, &y);
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), x, 0);
+}
+
+void
+gpdf_view_scroll_to_bottom (GPdfView *gpdf_view)
+{
+	gint x, y;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (gpdf_view), &x, &y);
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), x,
+				(GTK_LAYOUT (gpdf_view)->height
+				 - GTK_WIDGET (gpdf_view)->allocation.height));
+}
+
+gint
+gpdf_view_get_current_page (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+	
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), 0);
+	priv = gpdf_view->priv;
+	return priv->requested_page != priv->current_page ?
+	  priv->requested_page : priv->current_page; 
+}
+
 void
 gpdf_view_page_prev (GPdfView *gpdf_view)
 {
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
 	gpdf_view_goto_page (gpdf_view, gpdf_view->priv->requested_page - 1);
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 void
@@ -389,6 +627,7 @@ gpdf_view_page_next (GPdfView *gpdf_view)
 	g_return_if_fail(GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
 	gpdf_view_goto_page (gpdf_view, gpdf_view->priv->requested_page + 1);
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 void
@@ -397,6 +636,7 @@ gpdf_view_page_first (GPdfView *gpdf_view)
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
 	gpdf_view_goto_page (gpdf_view, 1);
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 void
@@ -406,13 +646,14 @@ gpdf_view_page_last (GPdfView *gpdf_view)
 
 	gpdf_view_goto_page (gpdf_view,
 			     gpdf_view->priv->pdf_doc->getNumPages ());
+	gpdf_view_scroll_to_top (gpdf_view);
 }
 
 gboolean
 gpdf_view_is_first_page (GPdfView *gpdf_view)
 {
 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), FALSE);
-	
+
 	return gpdf_view->priv->requested_page == 1;
 }
 
@@ -428,6 +669,129 @@ gpdf_view_is_last_page (GPdfView *gpdf_view)
 	return gpdf_view->priv->requested_page == last_page;
 }
 
+void
+gpdf_view_back_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+	GPdfViewHistory *hnode;
+	int target_page = 0;
+	double target_zoom = (double)0.0;
+	double target_hval, target_vval; 
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	priv = gpdf_view->priv;
+
+	g_return_if_fail (priv->backward_history != (G_List *)NULL);
+	g_return_if_fail (priv->backward_history->data != (gpointer)NULL);
+
+	hnode = (GPdfViewHistory *)priv->backward_history->data;
+	target_page = hnode->page;
+	target_zoom = hnode->zoom;
+	target_hval = hnode->hval; 
+	target_vval = hnode->vval; 
+	priv->backward_history = g_list_remove (priv->backward_history,
+						(gpointer)hnode);
+	if (!g_list_length (priv->backward_history)) {
+		g_list_free (priv->backward_history);
+		priv->backward_history = (G_List *)NULL;
+	}
+	hnode->page = priv->current_page;
+	hnode->zoom = priv->zoom;
+	hnode->hval = gtk_adjustment_get_value(GTK_LAYOUT (gpdf_view)->hadjustment); 
+	hnode->vval = gtk_adjustment_get_value(GTK_LAYOUT (gpdf_view)->vadjustment); 
+	priv->forward_history = g_list_prepend (priv->forward_history, hnode);
+
+	gpdf_view_goto_page_no_history (gpdf_view, target_page);
+	gpdf_view_zoom (gpdf_view, target_zoom, FALSE); 
+	gtk_adjustment_set_value(GTK_LAYOUT (gpdf_view)->hadjustment, target_hval);
+	gtk_adjustment_set_value(GTK_LAYOUT (gpdf_view)->vadjustment, target_vval);
+}
+
+void
+gpdf_view_forward_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+	GPdfViewHistory *hnode;
+	int target_page = 0;
+	double target_zoom = (double)0.0;
+	double target_hval, target_vval; 
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	priv = gpdf_view->priv;
+
+	g_return_if_fail (priv->forward_history != (G_List *)NULL);
+	g_return_if_fail (priv->forward_history->data != (gpointer)NULL);
+	hnode = (GPdfViewHistory *)priv->forward_history->data;
+	target_page = hnode->page;
+	target_zoom = hnode->zoom;
+	target_hval = hnode->hval; 
+	target_vval = hnode->vval; 
+	priv->forward_history = g_list_remove (priv->forward_history,
+					       (gpointer)hnode);
+	if (!g_list_length (priv->forward_history)) {
+		g_list_free (priv->forward_history);
+		priv->forward_history = (G_List *)NULL;
+	}
+	hnode->page = priv->current_page;
+	hnode->zoom = priv->zoom;
+	hnode->hval = gtk_adjustment_get_value(GTK_LAYOUT (gpdf_view)->hadjustment); 
+	hnode->vval = gtk_adjustment_get_value(GTK_LAYOUT (gpdf_view)->vadjustment); 
+	priv->backward_history = g_list_prepend (priv->backward_history, hnode);
+
+	gpdf_view_goto_page_no_history (gpdf_view, target_page);
+	gpdf_view_zoom (gpdf_view, target_zoom, FALSE); 
+	gtk_adjustment_set_value(GTK_LAYOUT (gpdf_view)->hadjustment, target_hval);
+	gtk_adjustment_set_value(GTK_LAYOUT (gpdf_view)->vadjustment, target_vval);
+}
+
+gboolean
+gpdf_view_is_first_history(GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), TRUE);
+	priv = gpdf_view->priv;
+
+	return (priv->backward_history == (G_List *)NULL);
+}
+
+gboolean
+gpdf_view_is_last_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), TRUE);
+	priv = gpdf_view->priv;
+
+	return (priv->forward_history == (G_List *)NULL);
+}
+
+static int
+gpdf_view_get_page_count (GgvDocument *document)
+{
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (document), -1);
+
+	return GPDF_VIEW (document)->priv->pdf_doc->getNumPages ();
+}
+
+static char **
+gpdf_view_get_page_names (GgvDocument *document)
+{
+	int page_count;
+	int i;
+	char **result;
+
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (document), NULL);
+
+	page_count = gpdf_view_get_page_count (document);
+
+	result = g_new0 (char *, page_count + 1);
+	for (i = 0; i < page_count; ++i)
+		result [i] = g_strdup_printf ("%d", i+1);
+
+	return result;
+}
+
 
 #define MAGSTEP  1.2
 #define IMAGSTEP 0.8333333333333
@@ -490,14 +854,11 @@ gpdf_view_zoom_out (GPdfView *gpdf_view)
 void
 gpdf_view_zoom_fit (GPdfView *gpdf_view, gint width, gint height)
 {
-	GPdfViewPrivate *priv;
 	gdouble x1, y1, x2, y2;
 	gdouble page_w, page_h, zoom_w, zoom_h;
 
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
-	priv = gpdf_view->priv;
-
 	gnome_canvas_get_scroll_region (GNOME_CANVAS (gpdf_view),
 					&x1, &y1, &x2, &y2);
 
@@ -516,14 +877,11 @@ gpdf_view_zoom_fit (GPdfView *gpdf_view, gint width, gint height)
 void
 gpdf_view_zoom_fit_width (GPdfView *gpdf_view, gint width)
 {
-	GPdfViewPrivate *priv;
 	gdouble x1, y1, x2, y2;
 	gdouble page_w;
 
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 
-	priv = gpdf_view->priv;
-
 	gnome_canvas_get_scroll_region (GNOME_CANVAS (gpdf_view),
 					&x1, &y1, &x2, &y2);
 	page_w = x2 - x1 + 0.5;
@@ -533,6 +891,59 @@ gpdf_view_zoom_fit_width (GPdfView *gpdf_view, gint width)
 	gpdf_view_zoom (gpdf_view, width / page_w, FALSE);
 }
 
+void
+gpdf_view_thumbnail_selected (GPdfView   *gpdf_view,
+			      gint x, gint y,
+			      gint w, gint h,
+			      gint page)
+{
+	gpdf_view_goto_page(gpdf_view, page);
+}
+
+void
+gpdf_view_annotation_selected(GPdfView *gpdf_view,
+			      Annot    *annot, 
+			      gint      page)
+{
+	gpdf_view_goto_page (gpdf_view, page);
+}
+
+void
+gpdf_view_annotation_toggled(GPdfView   *gpdf_view,
+			     Annot      *annot, 
+			     gint        page,
+			     gboolean	 active)
+{
+	if (gpdf_view->priv->current_page != page)
+	  gpdf_view_goto_page (gpdf_view, page);
+	else {
+		gpdf_view->priv->requested_page = page; 
+		gpdf_view->priv->current_page = 0; 
+		gpdf_view_render_page (gpdf_view);
+	}
+}
+
+void
+gpdf_view_bookmark_selected (GPdfView   *gpdf_view,
+			     LinkAction *link_action)
+{
+	g_return_if_fail (GPDF_IS_VIEW (gpdf_view));
+
+	switch (link_action->getKind ()) {
+	case actionGoTo:
+		gpdf_view_link_action_goto (
+			gpdf_view, dynamic_cast <LinkGoTo *> (link_action));
+		break;
+	case actionNamed:
+		gpdf_view_link_action_named (
+			gpdf_view, dynamic_cast <LinkNamed *> (link_action));
+		break;
+	default:
+		g_message ("Bookmark selection: unimplemented Action");
+		break;
+	}
+}
+
 
 static gboolean
 gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
@@ -581,7 +992,6 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Up:
 	case GDK_Up:
-		do_scroll = TRUE;
 		if (big_step) {
 			goto prev_screen;
 		} else {
@@ -591,7 +1001,6 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Down:
 	case GDK_Down:
-		do_scroll = TRUE;
 		if (big_step) {
 			goto next_screen;
 		} else {
@@ -608,36 +1017,38 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 	case GDK_KP_Delete:
 	case GDK_BackSpace:
 	prev_screen:
-		do_scroll = TRUE;
 		if (y <= 0) {
-			if (gpdf_view->priv->current_page > 1) {
-				gpdf_view_page_prev (gpdf_view);
-				/* 
-				 * FIXME height might be outdated for
-				 * new page
-				 */
-				y = GTK_LAYOUT (widget)->height - height;
-			}
+			if (gpdf_view->priv->current_page <= 1)
+				break;
+
+			gpdf_view_goto_page (
+				gpdf_view,
+				gpdf_view->priv->requested_page - 1);
+			gpdf_view_scroll_to_bottom (gpdf_view);
 		} else {
+			do_scroll = TRUE;
 			y -= height;
 		}
 		break;
 	case GDK_KP_Page_Down:
 	case GDK_Page_Down:
-		if (gpdf_view->priv->current_page < 
+		if (gpdf_view->priv->current_page <
 		    gpdf_view->priv->pdf_doc->getNumPages ())
 			gpdf_view_page_next (gpdf_view);
 		break;
 	case GDK_space:
 	next_screen:
-		do_scroll = TRUE;
 		if (y >= (gint)(GTK_LAYOUT (widget)->height) - height) {
-			if (gpdf_view->priv->current_page < 
-			    gpdf_view->priv->pdf_doc->getNumPages ()) {
-				gpdf_view_page_next (gpdf_view);
-				y = 0;
-			}
+			if (gpdf_view->priv->current_page >=
+			    gpdf_view->priv->pdf_doc->getNumPages ())
+				break;
+			
+			gpdf_view_goto_page (
+				gpdf_view,
+				gpdf_view->priv->requested_page + 1);
+			gpdf_view_scroll_to_top (gpdf_view);
 		} else {
+			do_scroll = TRUE;
 			y += height;
 		}
 		break;
@@ -688,12 +1099,13 @@ gpdf_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 		cursor = gdk_cursor_new (GDK_FLEUR);
 		gdk_pointer_grab (widget->window, FALSE,
 				  (GdkEventMask)
-				  (GDK_POINTER_MOTION_MASK | 
+				  (GDK_POINTER_MOTION_MASK |
 				   GDK_POINTER_MOTION_HINT_MASK |
 				   GDK_BUTTON_RELEASE_MASK),
 				  NULL, cursor, event->time);
 		gdk_cursor_unref (cursor);
 		return TRUE;
+		
 	default:
 		return FALSE;
 	}
@@ -707,13 +1119,13 @@ gpdf_view_button_release_event (GtkWidget *widget, GdkEventButton *event)
 	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
 
 	priv = GPDF_VIEW (widget)->priv;
-	
+
 	if (!priv->dragging || event->button != 1)
 		return FALSE;
 
 	priv->dragging = FALSE;
 	gdk_pointer_ungrab (event->time);
-	
+
 	return TRUE;
 }
 
@@ -739,7 +1151,7 @@ gpdf_view_handle_drag_motion (GPdfView *gpdf_view, GdkEventMotion *event)
 	dx = priv->drag_anchor_x - x;
 	dy = priv->drag_anchor_y - y;
 
-	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), 
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view),
 				priv->drag_offset_x + dx,
 				priv->drag_offset_y + dy);
 
@@ -755,7 +1167,7 @@ gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
 {
 
 	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
-	
+
 	if (GPDF_VIEW (widget)->priv->dragging)
 		gpdf_view_handle_drag_motion (GPDF_VIEW (widget), event);
 
@@ -815,6 +1227,7 @@ gpdf_view_set_pdf_doc (GPdfView *gpdf_view, PDFDoc *pdf_doc)
 
 	priv = gpdf_view->priv;
 
+	gpdf_view_clear_history (gpdf_view);
 	/*
 	 * All PDFDocs in GPdf code are owned by a GPdfPersistStream,
 	 * so don't delete our old pdf_doc
@@ -824,7 +1237,25 @@ gpdf_view_set_pdf_doc (GPdfView *gpdf_view, PDFDoc *pdf_doc)
 	priv->current_page = 0;
 	priv->output_dev->startDoc (pdf_doc->getXRef());
 	gpdf_view_page_background_show (gpdf_view);
-	gpdf_view_goto_page (gpdf_view, 1);
+	gpdf_view_goto_page_no_history (gpdf_view, 1);
+}
+
+/* Expects pathname in on-disk encoding (like Xpdf's saveAs) */
+void
+gpdf_view_save_as (GPdfView *gpdf_view, const gchar *pathname)
+{
+	GPdfViewPrivate *priv; 
+	GString *filename = new GString(pathname); 
+
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
+
+	priv = gpdf_view->priv;
+
+	gpdf_control_private_set_wait_cursor (priv->parent);
+	
+	priv->pdf_doc->saveAs (filename);
+	
+	gpdf_control_private_reset_cursor (priv->parent);
 }
 
 
@@ -836,12 +1267,12 @@ gpdf_view_dispose (GObject *object)
 	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (object));
 
 	priv = GPDF_VIEW (object)->priv;
-	
+
 	if (priv->output_dev) {
 		delete priv->output_dev;
 		priv->output_dev = NULL;
 	}
-	
+
 	if (priv->print_context) {
 		g_object_unref (priv->print_context);
 		priv->print_context = NULL;
@@ -856,6 +1287,12 @@ gpdf_view_dispose (GObject *object)
 		gtk_gesture_handler_destroy (priv->gesture_handler);
 		priv->gesture_handler = NULL;
 	}
+#ifdef USE_ANNOTS_VIEW
+	if (priv->annots_view) {
+		gtk_widget_unref (GTK_WIDGET (priv->annots_view));
+		priv->annots_view = NULL; 
+	}
+#endif
 
 	GNOME_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
 }
@@ -866,7 +1303,7 @@ gpdf_view_finalize (GObject *object)
 	GPdfView *gpdf_view;
 
 	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (object));
-	
+
 	gpdf_view = GPDF_VIEW (object);
 
 	if (gpdf_view->priv) {
@@ -928,12 +1365,29 @@ gpdf_view_class_init (GPdfViewClass *klass)
 		G_TYPE_NONE, 0);
 }
 
+static void
+ggv_document_interface_init (GgvDocumentClass *klass, gpointer iface_data)
+{
+	klass->get_page_count = gpdf_view_get_page_count;
+	klass->get_page_names = gpdf_view_get_page_names;
+}
+
 static void
 gpdf_view_instance_init (GPdfView *gpdf_view)
 {
 	gpdf_view->priv = g_new0 (GPdfViewPrivate, 1);
 }
 
+static void
+gpdf_view_setup_at (GPdfView *gpdf_view)
+{
+	AtkObject *accessible;
+
+	accessible = gtk_widget_get_accessible (GTK_WIDGET (gpdf_view));
+	atk_object_set_role (accessible, ATK_ROLE_DRAWING_AREA);
+	atk_object_set_name (accessible, _("Document View"));
+}
+
 static void
 gpdf_view_setup_gesture_handler (GPdfView *gpdf_view)
 {
@@ -993,8 +1447,10 @@ gpdf_view_setup_page_background (GPdfView *gpdf_view)
 	priv->page_shadow = gnome_canvas_item_new (
 		gnome_canvas_root (GNOME_CANVAS (gpdf_view)),
 		GNOME_TYPE_CANVAS_RECT,
-		"x1", PAGE_PAD - 1.0, "y1", (gdouble)PAGE_PAD - 1.0,
-		"x2", 72.0 + PAGE_PAD - 1.0, "y2", 72.0 + PAGE_PAD - 1.0,
+		"x1", (gdouble)PAGE_PAD -1.0,
+		"y1", (gdouble)PAGE_PAD -1.0,
+		"x2", 72.0 + (gdouble)PAGE_PAD -1.0,
+		"y2", 72.0 + (gdouble)PAGE_PAD -1.0,
 		"fill_color", "black", NULL);
 
 	gnome_canvas_item_lower_to_bottom (priv->page_shadow);
@@ -1011,24 +1467,71 @@ gpdf_view_construct (GPdfView *gpdf_view)
 
 	gpdf_view_setup_page_background (gpdf_view);
 	gpdf_view_setup_gesture_handler (gpdf_view);
+	gpdf_view_setup_at (gpdf_view);
 
 	priv = gpdf_view->priv;
 	priv->output_dev = new GPOutputDev ();
 	priv->zoom = 1.0;
 
+#ifdef USE_ANNOTS_VIEW
+	priv->annots_cb = gpdf_annots_view_is_annot_displayed;
+#endif
+	
 	return gpdf_view;
 }
 
 GtkWidget *
-gpdf_view_new (void)
+gpdf_view_new (GPdfControl *control)
 {
 	GPdfView *gpdf_view;
 
 	gpdf_view = GPDF_VIEW (g_object_new (GPDF_TYPE_VIEW,
 					     "aa", TRUE,
 					     NULL));
+	gpdf_view->priv->parent = control; 
 
 	return GTK_WIDGET (gpdf_view_construct (gpdf_view));
 }
 
+static void
+gpdf_view_class_init_trampoline (gpointer klass, gpointer data)
+{
+	parent_class = (GnomeCanvasClass *)g_type_class_ref (PARENT_TYPE);
+	gpdf_view_class_init ((GPdfViewClass *)klass);
+}
+
+GType
+gpdf_view_get_type (void)
+{
+	static GType object_type = 0;
+	if (object_type == 0) {
+		static const GTypeInfo object_info = {
+		    sizeof (GPdfViewClass),
+		    NULL,		/* base_init */
+		    NULL,		/* base_finalize */
+		    gpdf_view_class_init_trampoline,
+		    NULL,		/* class_finalize */
+		    NULL,               /* class_data */
+		    sizeof (GPdfView),
+		    0,                  /* n_preallocs */
+		    (GInstanceInitFunc) gpdf_view_instance_init
+		};
+
+		static const GInterfaceInfo ggv_document_info = {
+			(GInterfaceInitFunc) ggv_document_interface_init,
+			NULL,
+			NULL
+		};
+
+		object_type = g_type_register_static (PARENT_TYPE, "GPdfView",
+						      &object_info,
+						      (GTypeFlags)0);
+
+		g_type_add_interface_static (object_type,
+					     GGV_DOCUMENT_TYPE,
+					     &ggv_document_info);
+	}
+	return object_type;
+}
+
 END_EXTERN_C
